@@ -3,6 +3,7 @@ import socket
 import logging
 from contextlib import closing
 from collections import (OrderedDict, namedtuple)
+
 import substance.core
 from substance.monads import *
 from substance.config import (Config)
@@ -10,10 +11,14 @@ from substance.logs import *
 from substance.shell import Shell
 from substance.link import Link
 from substance.box import Box
-from substance.utils import mergeDict, parseDotEnv
+from substance.utils import mergeDict, mergeDictOverwrite, parseDotEnv, CommandList, Memoized
+from substance.path import (getHomeDirectory, inner)
 from substance.hosts import SubHosts
 from substance.driver.virtualbox import VirtualBoxDriver
-from substance.constants import (EngineStates)
+from substance.constants import (EngineStates, DefaultEngineBox, Orchestrators)
+from substance.platform import (isWithinWindowsSubsystem)
+from substance.orchestrator import (Dockwrkr, Compose)
+
 from substance.exceptions import (
     FileSystemError,
     ConfigValidationError,
@@ -23,7 +28,8 @@ from substance.exceptions import (
     EngineNotProvisioned,
     EngineProvisioned,
     EnvNotDefinedError,
-    InvalidCommandError
+    InvalidCommandError,
+    InvalidOrchestrator
 )
 from substance.syncher.unison import UnisonSyncher
 
@@ -78,13 +84,14 @@ class Engine(object):
         self.config = Config(configFile)
         self.logAdapter = EngineLogAdapter(logger, self.__dict__)
         self.currentEnv = None
+        self.subenvConfig = None
 
     def getDefaults(self):
         defaults = OrderedDict()
         defaults['name'] = 'default'
         defaults['driver'] = 'virtualbox'
         defaults['id'] = None
-        defaults['box'] = 'turbulent/substance-box:0.6'
+        defaults['box'] = DefaultEngineBox
         defaults['profile'] = EngineProfile().__dict__
         defaults['docker'] = OrderedDict()
         defaults['docker']['port'] = 2375
@@ -96,7 +103,7 @@ class Engine(object):
         defaults['network']['sshIP'] = None
         defaults['network']['sshPort'] = 4500
         defaults['devroot'] = OrderedDict()
-        defaults['devroot']['path'] = os.path.join('~', 'devroot')
+        defaults['devroot']['path'] = getHomeDirectory('devroot')
         defaults['devroot']['mode'] = 'unison'
         defaults['devroot']['syncArgs'] = [
             '-ignore', 'Path */var',
@@ -132,7 +139,7 @@ class Engine(object):
         ops.append(self.confValidateDevroot(config.get('devroot', {})))
 
         def dd(err):
-            print("%s" % err)
+            print(("%s" % err))
             return ConfigValidationError(err.message)
 
         return Try.sequence(ops) \
@@ -237,6 +244,12 @@ class Engine(object):
     def loadConfigFile(self):
         return self.config.loadConfigFile().bind(self.validateConfig).map(self.chainSelf)
 
+    def loadSubenvConfigFile(self):
+        basePath = os.path.expanduser(self.config.get('devroot').get('path'))
+        self.subenvConfig = Config(os.path.join(basePath,
+                                   '%s/.substance/subenv.yml' % self.currentEnv))
+        return self.subenvConfig.loadConfigFile().map(self.chainSelf)
+
     def loadState(self):
         ddebug("Engine load state %s" % self.name)
         return self.fetchState().bind(self.setState).map(self.chainSelf)
@@ -285,7 +298,7 @@ class Engine(object):
             cf['profile']['cpus'] = profile.cpus
             cf['profile']['memory'] = profile.memory
 
-        for ck, cv in cf.iteritems():
+        for ck, cv in cf.items():
             self.config.set(ck, cv)
 
         return OK(self.config)
@@ -512,14 +525,14 @@ class Engine(object):
         hostsCmd = "sed -i \"s/127.0.1.1.*/127.0.1.1\t%s/g\" /etc/hosts" % self.name
         # hostsCmd = "hostnamectl set-hostname %s" % dns
         serviceCmd = "service hostname restart"
-        cmd = "sudo -- bash -c '%s && %s && %s && %s'" % (
+        cmd = "sudo -- /bin/sh -c '%s && %s && %s && %s'" % (
             echoCmd, hostsCmd, hostCmd, serviceCmd)
-        cmds = map(defer(self.link.runCommand, stream=True, sudo=False), [cmd])
+        cmds = list(map(defer(self.link.runCommand, stream=True, sudo=False), [cmd]))
         return Try.sequence(cmds)
 
     def exposePort(self, local_port, public_port, scheme):
         keyfile = self.core.getInsecureKeyFile()
-        keyfile = Shell.normalizePath(keyfile)
+        keyfile = inner(keyfile)
         forward_descr = "0.0.0.0:%s:127.0.0.1:%s" % (public_port, local_port)
         engineIP = self.getSSHIP()
         enginePort = str(self.getSSHPort())
@@ -527,7 +540,7 @@ class Engine(object):
         cmdArgs = ["ssh", "-N", "-L", forward_descr, engineIP, "-l", "substance", "-p", enginePort,
                    "-o", "StrictHostKeyChecking=no", "-o", "UserKnownHostsFile=/dev/null", "-i", keyfile]
         sudo = False
-        if public_port < 1024 and not Shell.isCygwin():
+        if public_port < 1024 and not isWithinWindowsSubsystem():
             sudo = True
         self.logAdapter.info(
             "Exposing port %s as %s; kill the process (CTRL-C) to un-expose.", local_port, public_port)
@@ -542,21 +555,61 @@ class Engine(object):
                     "Others can now connect to port %s via IP %s.", public_port, ip)
         Shell.execvp(cmdPath, cmdArgs, {}, sudo=sudo)
 
+    def getOrchestrator(self):
+        if self.subenvConfig:
+            return self._getOrchestrator()
+
+        return self.envLoadCurrent() \
+            .then(self.loadSubenvConfigFile) \
+            .catch(lambda err: OK(None)) \
+            .then(self._getOrchestrator)
+
+    def _getOrchestrator(self):
+        orchestrator = self.subenvConfig.get('orchestrator') if self.subenvConfig else None
+        logger.debug("Orchestrator: %s" % orchestrator)
+        if orchestrator and orchestrator == Orchestrators.COMPOSE:
+            return OK(Compose(self))
+        elif orchestrator and orchestrator != Orchestrators.DOCKWRKR:
+            return Fail(InvalidOrchestrator('%s is not a known orchestrator' % orchestrator))
+        else:
+            return OK(Dockwrkr(self))
+
+    def getOrchestratorCommand(self, command, *args):
+        orch = self.getOrchestrator()
+        if orch.isFail():
+            raise orch.value
+        else:
+            orch = orch.getOK()
+
+        if orch:
+            return getattr(orch, command)(*args)
+        else:
+            return ''
+
+    def orchestrate(self, command, *args):
+        commands = self.getOrchestratorCommand(command, *args)
+        if len(commands) > 0:
+            return list(map(lambda x: "(subenv run %s)" % x, commands))
+
     def envSwitch(self, subenvName, restart=False):
         self.logAdapter.info(
             "Switch engine '%s' to subenv '%s'" % (self.name, subenvName))
-        cmds = [
+
+        self.currentEnv = subenvName
+        self.loadSubenvConfigFile()
+
+        cmds = CommandList([
             "subenv init '/substance/devroot/%s'" % (subenvName),
             "subenv use '%s'" % (subenvName),
-        ]
+        ])
 
         if restart:
-            cmds.append("dockwrkr -y reset")
-            cmds.append("subenv run dockwrkr login")
-            cmds.append("subenv run dockwrkr start -a")
+            cmds.append(self.orchestrate('resetAll'))
+            cmds.append(self.orchestrate('login'))
+            cmds.append(self.orchestrate('startAll'))
 
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, sudo=False) \
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False) \
             .then(self.envRegister)
 
     def envRegister(self):
@@ -576,58 +629,57 @@ class Engine(object):
             return SubHosts.register(env['SUBENV_NAME'] + self.core.config.get('tld'), self.getPublicIP())
 
     def envLoadCurrent(self):
-        cmds = ["subenv current"]
+        cmds = CommandList(["subenv current"])
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=False, sudo=False, capture=True) \
+            .bind(Link.runCommand, cmds.logicAnd(), stream=False, sudo=False, capture=True) \
             .map(self.__cacheCurrentEnv) \
             .catch(lambda err: Fail(EnvNotDefinedError("No current subenv is set. Check 'switch' for detais."))) \
             .then(lambda: self)
 
     def envStart(self, reset=False, containers=[]):
-        cmds = []
+        cmds = CommandList()
+
         if reset:
-            cmds.append("dockwrkr -y reset")
+            cmds.append(self.getOrchestratorCommand('resetAll'))
 
         if len(containers) > 0:
             self.logAdapter.info("Starting %s container(s)" %
                                  (' '.join(containers)))
-            cmds.append('subenv run dockwrkr start %s' % ' '.join(containers))
+            cmds.append(self.orchestrate('start', containers))
         else:
             self.logAdapter.info("Starting all containers...")
-            cmds.append('subenv run dockwrkr start -a')
+            cmds.append(self.orchestrate('startAll'))
 
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)
 
     def envRestart(self, time=10, containers=[]):
-        cmds = []
+        cmds = CommandList()
 
         if len(containers) > 0:
             self.logAdapter.info("Restarting %s containers" %
                                  (' '.join(containers)))
-            cmds.append('subenv run dockwrkr restart -t %s %s' %
-                        (time, ' '.join(containers)))
+            cmds.append(self.orchestrate('restart', containers, time))
         else:
             self.logAdapter.info("Restarting all containers...")
-            cmds.append('subenv run dockwrkr restart -a -t %s' % time)
+            cmds.append(self.orchestrate('restartAll', time))
 
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)
 
     def envRecreate(self, time=10, containers=[]):
-        cmds = []
+        cmds = CommandList()
 
         if len(containers) > 0:
             self.logAdapter.info("Recreating %s containers" %
                                  (' '.join(containers)))
-            cmds.append('subenv run dockwrkr recreate -t %s %s' %
-                        (time, ' '.join(containers)))
+            cmds.append(self.orchestrate('recreate', containers, time))
         else:
             self.logAdapter.info("Recreating all containers...")
-            cmds.append('subenv run dockwrkr recreate -a -t %s' % time)
+            cmds.append(self.orchestrate('recreateAll', time))
 
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)
 
     def envShell(self, container=None, user=None, cwd=None):
         if container:
@@ -640,30 +692,19 @@ class Engine(object):
         return self.envExec(container, ["exec /bin/bash"], cwd, user)
 
     def envExec(self, container, cmd, cwd=None, user=None):
-        opts = []
-        cmdline = ' '.join(cmd)
-
-        initCommands = ['export TERM=xterm', cmdline]
-        if user:
-            opts.append('--user %s' % user)
-        if cwd:
-            initCommands.insert(0, 'cd %s' % cwd)
-
-        tpl = {
-            'opts': ' '.join(opts),
-            'container': container,
-            'initCommands': ' && '.join(initCommands)
-        }
-
-        cmd = "subenv run dockwrkr exec -t -i %(opts)s %(container)s 'bash -c \"%(initCommands)s\"'" % tpl
-        return self.readLink().bind(Link.runCommand, cmd=cmd, interactive=True, stream=True, shell=False, capture=False)
+        cmd = CommandList(self.orchestrate('exec', container, cmd, cwd, user))
+        return self.readLink().bind(Link.runCommand, cmd=cmd.logicAnd(), interactive=True, stream=True, shell=False, capture=False)
 
     def envRun(self, task, args):
-        cmd = "subenv run dockwrkr run %s %s" % ( task, ' '.join(args))
-        return self.readLink().bind(Link.runCommand, cmd=cmd, interactive=True, stream=True, shell=False, capture=False)
+        cmd = CommandList(self.orchestrate('run', task, args))
+        return self.readLink().bind(Link.runCommand, cmd=cmd.logicAnd(), interactive=True, stream=True, shell=False, capture=False)
 
     def envExecAlias(self, alias, args):
-        aliases = self.config.get('aliases')
+        aliases = self.config.get('aliases') if self.config.get('aliases') else {}
+        subenvAliases = self.subenvConfig.get('aliases') if self.subenvConfig.get('aliases') else {}
+        if subenvAliases:
+            aliases = mergeDictOverwrite(aliases, subenvAliases)
+
         if aliases and alias in aliases:
             cmd = aliases[alias]
             args = cmd['args'] + args
@@ -671,41 +712,41 @@ class Engine(object):
         return Fail(InvalidCommandError("Invalid command '%s' specified for '%s'.\n\nUse 'substance help' for available commands." % (alias, 'substance')))
 
     def envStop(self, time=10, containers=[]):
-        cmds = []
+        cmds = CommandList()
+
         if len(containers) > 0:
             self.logAdapter.info("Stopping %s container(s)" %
                                  (' '.join(containers)))
-            cmds.append('subenv run dockwrkr stop -t %s %s' %
-                        (time, ' '.join(containers)))
+
+            cmds.append(self.orchestrate('stop', containers, time))
         else:
             self.logAdapter.info("Stopping containers...")
-            cmds.append('subenv run dockwrkr stop -a -t %s' % time)
+            cmds.append(self.orchestrate('stopAll', time))
 
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)
 
     def envStatus(self, full=False):
         if full:
-            cmds = []
-            cmds.append('subenv run dockwrkr status')
+            cmds = CommandList([self.orchestrate('status')])
             return self.readLink() \
-                .bind(Link.runCommand, ' && '.join(cmds), stream=False, sudo=False, capture=True) \
+                .bind(Link.runCommand, cmds.logicAnd(), stream=False, sudo=False, capture=True) \
                 .map(self.__envStatus)
         else:
             return OK(self.__envStatus())
 
     def envCleanup(self):
-        cmds = []
+        cmds = CommandList()
         cmds.append('docker system prune -a')
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), interactive=True, stream=True, shell=False, capture=False)
+            .bind(Link.runCommand, cmds.logicAnd(), interactive=True, stream=True, shell=False, capture=False)
 
     def envLogs(self, parts=[], pattern=None, follow=True, lines=None):
 
         if not pattern:
             pattern = "%s*.log" % ('-'.join(parts))
 
-        cmds = []
+        cmds = CommandList()
         cmd = "tail"
         if follow:
             cmd += " -f"
@@ -716,19 +757,19 @@ class Engine(object):
 
         cmds.append('subenv run %s' % cmd)
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, capture=False, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, capture=False, sudo=False)
 
     def envListLogs(self, parts=[], pattern=None):
 
         if not pattern:
             pattern = "%s*.log" % ('-'.join(parts))
 
-        cmds = []
+        cmds = CommandList()
         cmd = "ls -1 \"logs/%s\" | xargs -n1 basename" % pattern
 
         cmds.append('subenv run %s' % cmd)
         return self.readLink() \
-            .bind(Link.runCommand, ' && '.join(cmds), stream=True, capture=False, sudo=False)
+            .bind(Link.runCommand, cmds.logicAnd(), stream=True, capture=False, sudo=False)
 
     def __envStatus(self, containers=None):
         return {
@@ -760,7 +801,7 @@ class Engine(object):
     def mountFolders(self):
         self.logAdapter.info("Mounting engine folders")
         folders = self.getEngineFolders()
-        return Try.of(map(self.mountFolder, folders))
+        return Try.of(list(map(self.mountFolder, folders)))
 
     def mountFolder(self, folder):
         mountCmd = "mount -t vboxsf -o umask=%(umask)s,gid=%(gid)s,uid=%(uid)s %(name)s %(guestPath)s" % folder.__dict__
