@@ -9,20 +9,20 @@ from collections import namedtuple
 from substance.monads import *
 from substance.exceptions import *
 from subprocess import check_output
-from substance.termutils import getTerminalSize
+from substance.termutils import (
+    hasTermios,
+    saveTerminalAttrs,
+    restoreTerminalAttrs,
+    setTerminalInteractive,
+    getTerminalSize,
+    decodeTerminalBytes
+)
 
 logger = logging.getLogger(__name__)
 
 LinkResponse = namedtuple(
     'LinkResponse', ['link', 'stdin', 'stdout', 'stderr', 'code', 'cmd'])
 
-try:
-    import termios
-    import tty
-    hasTermios = True
-except ImportError:
-    hasTermios = False
-
 
 class Link(object):
 
@@ -139,7 +139,7 @@ class Link(object):
         logger.debug("LINKCOMMAND: %s" % cmd)
 
         try:
-            oldtty = termios.tcgetattr(sys.stdin)
+            saveTerminalAttrs(sys.stdin)
 
             # Setup our connection channel with forwarding
             channel = None
@@ -152,16 +152,15 @@ class Link(object):
 
             if interactive:
                 channel.settimeout(0.0)
-                tty.setraw(sys.stdin.fileno())
-                tty.setcbreak(sys.stdin.fileno())
+                setTerminalInteractive(sys.stdin)
 
             forward = paramiko.agent.AgentRequestHandler(channel)
 
             # Setup our buffers
             bufsize = 1024
-            stdin = ''
-            stdout = ''
-            stderr = ''
+            stdin =  b''
+            stdout = b''
+            stderr = b''
 
             # Star the command stream
             if shell:
@@ -182,46 +181,49 @@ class Link(object):
                     isAlive = False
                     break
 
-                if channel in r:
-                    if channel.recv_ready():
-                        d = channel.recv(bufsize)
-                        if len(d) == 0:
-                            isAlive = False
-                        else:
-                            if stream:
-                                sys.stdout.write(d)
-                                sys.stdout.flush()
-                            if capture:
-                                stdout += d
-
-                    if channel.recv_stderr_ready():
-                        d = channel.recv_stderr(bufsize)
+                if channel in r and channel.recv_ready():
+                    d = channel.recv(bufsize)
+                    if len(d) == 0:
+                        isAlive = False
+                    else:
                         if stream:
-                            sys.stderr.write(d)
-                            sys.stderr.flush()
+                            sys.stdout.write(decodeTerminalBytes(d))
+                            sys.stdout.flush()
                         if capture:
-                            stderr += d
+                            stdout += d
+
+                if channel in r and channel.recv_stderr_ready():
+                    d = channel.recv_stderr(bufsize)
+                    if stream:
+                        sys.stderr.write(decodeTerminalBytes(d))
+                        sys.stderr.flush()
+                    if capture:
+                        stderr += d
 
                 if sys.stdin in r and isAlive and interactive:
                     x = os.read(sys.stdin.fileno(), bufsize)
                     if len(x) == 0:
                         isAlive = False
                     else:
-                        channel.send(x)
+                        channel.send(decodeTerminalBytes(x))
 
             code = channel.recv_exit_status()
+            stdout = decodeTerminalBytes(stdout)
+            stdin = decodeTerminalBytes(stdin)
+            stderr = decodeTerminalBytes(stderr)
 
             if code != 0:
+                logger.debug("code %s, stdout: %s, stderr: %s" % (code, stdout, stderr))
                 return Fail(LinkCommandError(cmd=cmd, message="An error occured when running command.", stdout=stdout, stderr=stderr, code=code, link=self))
 
             return OK(LinkResponse(link=self, cmd=cmd, stdin=stdin, stdout=stdout, stderr=stderr, code=code))
-        except Exception as err:
-            logger.debug("%s" % err)
-            return Fail(err)
+        # except Exception as err:
+        #     logger.debug("%s" % err.stdout)
+        #     return Fail(err)
         except KeyboardInterrupt as err:
             return Fail(LinkCommandError(cmd=cmd, message="User Interrupted.", stdout=stdout, stderr=stderr, code=1, link=self))
         finally:
-            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldtty)
+            restoreTerminalAttrs(sys.stdin)
 
     def resizePTY(self, channel):
         tty_width, tty_height = getTerminalSize()
@@ -266,7 +268,7 @@ class Link(object):
 
         logger.debug("Opening interactive POSIX shell")
         sys.stdout.write('\r\n*** Begin interactive session.\r\n')
-        oldtty = termios.tcgetattr(sys.stdin)
+        saveTerminalAttrs(sys.stdin)
 
         channel = self.client.get_transport().open_session()
         forward = paramiko.agent.AgentRequestHandler(channel)
@@ -279,8 +281,7 @@ class Link(object):
             channel.send("\n")
 
         try:
-            tty.setraw(sys.stdin.fileno())
-            tty.setcbreak(sys.stdin.fileno())
+            setTerminalInteractive(sys.stdin)
             channel.settimeout(0.0)
 
             isAlive = True
@@ -294,7 +295,7 @@ class Link(object):
                         if len(x) == 0:
                             isAlive = False
                         else:
-                            sys.stdout.write(x)
+                            sys.stdout.write(decodeTerminalBytes(x))
                             sys.stdout.flush()
                     except socket.timeout:
                         pass
@@ -307,7 +308,7 @@ class Link(object):
             channel.shutdown(2)
 
         finally:
-            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldtty)
+            restoreTerminalAttrs(sys.stdin)
             sys.stdout.write('\r\n*** End of interactive session.\r\n')
 
     def _windowsShell(self, cmd=None):
@@ -323,11 +324,12 @@ class Link(object):
                     sys.stdout.write('\r\n*** EOF ***\r\n\r\n')
                     sys.stdout.flush()
                     break
-                sys.stdout.write(data)
+                sys.stdout.write(decodeTerminalBytes(data))
                 sys.stdout.flush()
 
         writer = threading.Thread(target=writeall, args=(chan,))
         writer.start()
+
         try:
             while True:
                 d = sys.stdin.read(1)
