@@ -5,10 +5,12 @@ import tempfile
 import shutil
 import logging
 import collections
+import functools
 import os
 import errno
 import hashlib
 import tarfile
+from math import floor
 from time import time
 from pkg_resources import Requirement, resource_filename, require
 from substance.exceptions import (
@@ -23,7 +25,7 @@ _yaml_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
 
 
 def _dict_representer(dumper, data):
-    return dumper.represent_dict(data.iteritems())
+    return dumper.represent_dict(iter(data.items()))
 
 
 def _dict_constructor(loader, node):
@@ -34,10 +36,10 @@ yaml.add_representer(collections.OrderedDict, _dict_representer)
 yaml.add_constructor(_yaml_mapping_tag, _dict_constructor)
 
 
-def writeToFile(filename, data):
+def writeToFile(filename, the_bytes):
     try:
         with open(filename, "wb") as fh:
-            fh.write(data)
+            fh.write(the_bytes)
     except Exception as err:
         raise FileSystemError("Failed to write to %s : %s" % (filename, err))
 
@@ -52,10 +54,10 @@ def writeYAML(filename, data):
 
 def readYAML(filename):
     try:
-        stream = open(filename, "r")
-        contents = yaml.load(stream)
-        return contents
-    except yaml.YAMLError, exc:
+        with open(filename, "r") as stream:
+            contents = yaml.load(stream, Loader=yaml.SafeLoader)
+            return contents
+    except yaml.YAMLError as exc:
         msg = "Syntax error in file %s"
         if hasattr(exc, 'problem_mark'):
             mark = exc.problem_mark
@@ -101,7 +103,7 @@ def streamDownload(url, destination):
             else:
                 progress = 0
                 startTime = time()
-                chunkSize = contentLength / 100
+                chunkSize = floor(contentLength / 100)
                 for chunk in r.iter_content(chunk_size=chunkSize):
                     if chunk:
                         progress += len(chunk)
@@ -162,23 +164,6 @@ def untar(filename, destination=None):
     return filename
 
 
-def pathComponents(path):
-    folders = []
-    path = os.path.normpath(path)
-    while 1:
-        path, folder = os.path.split(path)
-
-        if folder != "":
-            folders.append(folder)
-        else:
-            if path != "":
-                folders.append(path)
-            break
-
-    folders.reverse()
-    return folders
-
-
 def mergeDict(a, b, path=None):
     if path is None:
         path = []
@@ -195,6 +180,13 @@ def mergeDict(a, b, path=None):
     return a
 
 
+# merge 2 dicts and overwrite target keys with source keys if collision
+def mergeDictOverwrite(target, source):
+    merged = target.copy()
+    merged.update(source)
+    return merged
+
+
 def readDotEnv(filepath, env={}):
     with open(filepath) as f:
         return parseDotEnv(f, env)
@@ -214,7 +206,7 @@ def parseDotEnv(dotenvdata, env={}):
 def makeSymlink(source, link, force=False):
     try:
         os.symlink(source, link)
-    except OSError, e:
+    except OSError as e:
         if e.errno == errno.EEXIST and force:
             os.remove(link)
             os.symlink(source, link)
@@ -226,9 +218,59 @@ def readSymlink(link):
     return os.readlink(link)
 
 
-def expandLocalPath(path, basePath=None):
-    path = os.path.expanduser(path)
-    if not os.path.isabs(path) and basePath:
-        path = os.path.normpath(path)
-        path = os.path.join(basePath, path)
-    return os.path.abspath(os.path.realpath(path))
+def flatten(l):
+    for el in l:
+        if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
+            yield from flatten(el)
+        else:
+            yield el
+
+
+class CommandList(list):
+    def commands(self):
+        cmds = list(filter(lambda x: x is not None, self))
+        return flatten(cmds)
+
+    def joined(self, by=' '):
+        return by.join(self.commands())
+
+    def logicAnd(self):
+        return self.joined(' && ')
+
+    def logicOr(self):
+        return self.joined(' || ')
+
+
+class Memoized(object):
+    '''Decorator. Caches a function's return value each time it is called.
+    If called later with the same arguments, the cached value is returned
+    (not reevaluated).
+    '''
+    def __init__(self, func):
+        self.func = func
+        self.cache = {}
+
+    def __call__(self, *args):
+        if not isinstance(args, collections.Hashable):
+            # uncacheable. a list, for instance.
+            # better to not cache than blow up.
+            return self.func(*args)
+        if args in self.cache:
+            return self.cache[args]
+        else:
+            value = self.func(*args)
+            self.cache[args] = value
+            return value
+
+    def __repr__(self):
+        '''Return the function's docstring.'''
+        return self.func.__doc__
+
+    def __get__(self, obj, objtype):
+        '''Support instance methods.'''
+        return functools.partial(self.__call__, obj)
+
+
+class DictCompat(dict):
+    def iteritems(self):
+        return self.items()
